/*
 * Copyright (c) 2012,2016 Anthony Mallet
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                                        Sat Aug  4 2012
 */

/* This hooks on the atexit(3) function, because some inelegant plugins may
 * want to register atexit handlers (e.g., if the plugin does call ros::init()
 * from the ros library). Of course, this will crash badly during program exit,
 * because plugins are dlclose()'d before program termination, so the
 * registered handlers are dandling in the void when they are called by exit().
 *
 * This is a well known problem and the gnu libc or Solaris, for instance,
 * provide a special atexit() that recognize this situation. Not everybody
 * uses those C library, however. An alternative solution could be to call
 * _exit(0) instead of returning from main(), but this would not call
 * legitimate atexit() handlers (e.g., the rtld cleanup, C++ dtors...) and may
 * raise other unexpected issues?
 *
 * The fix here seem the less bad workaround, since it does nothing explicitly
 * non-portable. Does it?
 */

/* Usage is as follow:
 *
 * - by default, this atexit() does nothing but calling the original one.
 *
 * - when calling code in a plugin, wrap the call in question within a
 *   gnmx_atexit_push(dlhandle)/gnmx_atexit_pop() pair, where dlhandle is the
 *   handle obtained when dlopen()'ing the plugin. This will make atexit()
 *   register the handler for that plugin. Then, call
 *   gnmx_atexit_invoke(dlhandle) right before dlclose()'ing the plugin. This
 *   will invoke (and delete) the handlers for that plugin.
 */
#include "acgenomix.h"

#include <assert.h>
#include <dlfcn.h>
#include <err.h>
#include <stdlib.h>

/* genomix does not use threads, but plugins do... */
#ifdef HAVE_PTHREAD
# include <pthread.h>
#endif


/* --- local data ---------------------------------------------------------- */

struct gnmx_atexit_handler {
  void *dso;
  void (*handler)(void);
  struct gnmx_atexit_handler *next;
};

static void *gnmx_atexit_dso = NULL;

/* stack of handlers */
static struct gnmx_atexit_handler *gnmx_atexit_stack = NULL;

#ifdef HAVE_PTHREAD
/* a mutex for _REENTRANT contexts */
static pthread_mutex_t gnmx_atexit_mutex = PTHREAD_MUTEX_INITIALIZER;

# define gnmx_atexit_lock	pthread_mutex_lock(&gnmx_atexit_mutex)
# define gnmx_atexit_unlock	pthread_mutex_unlock(&gnmx_atexit_mutex)
#else
# define gnmx_atexit_lock
# define gnmx_atexit_unlock
#endif


/* --- atexit -------------------------------------------------------------- */

int
atexit(void (*handler)())
{
  static int (*atexit_real)(void (*h)()) = NULL;

  struct gnmx_atexit_handler *h;

  /* The first time we are called, lookup the original atexit(3). If it cannot
   * be found, still mark it as initialized by pointing it on ourself. */
  gnmx_atexit_lock;
  if (!atexit_real) {
    atexit_real = dlsym(RTLD_NEXT, "atexit");
    if (!atexit_real) {
      warnx("%s", dlerror());
      atexit_real = atexit;
    }
  }
  assert(atexit_real);

  /* If we are called from the main program, use the original atexit(3) unless
   * it could not be resolved. In this case, fall back to private registration
   */
  if (!gnmx_atexit_dso && atexit_real != atexit) {
    gnmx_atexit_unlock;
    return atexit_real(handler);
  }

  /* We are called from within a dso, register the handler privately */
  warnx("a plugin is using atexit()");

  h = malloc(sizeof(*h));
  if (!h) {
    gnmx_atexit_unlock;
    return -1;
  }

  h->dso = gnmx_atexit_dso;
  h->handler = handler;
  assert(h->handler);

  h->next = gnmx_atexit_stack;
  gnmx_atexit_stack = h;

  gnmx_atexit_unlock;
  return 0;
}


/* --- gnmx_atexit_push/pop ------------------------------------------------ */

/* This is not thread safe, but in this program this is called from a single
 * thread. */

void
gnmx_atexit_push(void *dlhandle, void **prev)
{
  assert(dlhandle);
  *prev = gnmx_atexit_dso;
  gnmx_atexit_dso = dlhandle;
}

void
gnmx_atexit_pop(void *prev)
{
  gnmx_atexit_dso = prev;
}


/* --- gnmx_atexit_invoke -------------------------------------------------- */

void
gnmx_atexit_invoke(void *dlhandle)
{
  struct gnmx_atexit_handler *p, *s;
  void *ath;

  do {
    /* Unlink the first handler matching dlhandle */
    gnmx_atexit_lock;
    if (!gnmx_atexit_stack) {
      gnmx_atexit_unlock;
      break;
    }
    if (gnmx_atexit_stack->dso == dlhandle) {
      s = gnmx_atexit_stack;
      gnmx_atexit_stack = s->next;
    } else {
      for (s = NULL, p = gnmx_atexit_stack; p->next; p = p->next)
        if (p->next->dso == dlhandle) {
          s = p->next;
          p->next = s->next;
          break;
        }
    }
    gnmx_atexit_unlock;
    if (!s) break;

    /* Run it in the dlhandle context (in case this adds other handlers) */
    if (dlhandle) gnmx_atexit_push(dlhandle, &ath);
    s->handler();
    if (dlhandle) gnmx_atexit_pop(ath);
    free(s);

    /* Start over */
  } while(1);
}
